using UnityEngine;

/// <summary>
/// <see cref="Fact.Id"/> <c>MonoBehaviour</c> wrapper to be attached to a <see cref="GameObject"/>
/// </summary>
public class FactWrapper : MonoBehaviour//, ISerializationCallbackReceiver
{
    public enum FactMaterials
    {
        Default = 0,
        Selected = 1,
        Hint = 2,
        Solution = 3,
    }

    /// <summary>
    /// <see cref="Fact.Id"/> to identify arbitrary <see cref="Fact"/> by its <see cref="Fact.WorldRepresentation"/>
    /// </summary>
    public string URI
    {
        get => _URI;
        set
        {
            _Fact = null;
            _URI = value;
            FactUpdated();
        }
    }
    [SerializeField] // So cloning takes over HideInInspector, 
    protected string _URI;

    /// <summary>
    /// Shorthand and Cache for <see cref="FactRecorder.AllFacts.TryGetValue(string, out Fact)"/>.
    /// </summary>
    public Fact Fact
    {
        get
        {
            if (_Fact == null && URI != null)
                FactRecorder.AllFacts.TryGetValue(URI, out _Fact);
            return _Fact;
        }
        set
        {
            _Fact = value;
            _URI = value?.Id;
            FactUpdated();
        }
    }
    protected Fact _Fact;

    #region UnityMethods
    // region UnityMethods:
    // protected not private: Note to Derived to not implement

    protected void Awake()
    {
        _Awake();
    }
    protected virtual void _Awake() { }

    protected void Start()
    {
        _Start();
    }
    protected virtual void _Start() { }

    protected void OnEnable()
    {
        //CommunicationEvents.AnimateExistingFactEvent.RemoveAllListeners();
        CommunicationEvents.AnimateExistingFactEvent.AddListener(HiglightFactEvent);
        CommunicationEvents.RemoveFactEvent.AddListener(DeleteFactEvent);
        _OnEnable();
    }
    protected virtual void _OnEnable() { }

    protected void OnDisable()
    {
        CommunicationEvents.AnimateExistingFactEvent.RemoveListener(HiglightFactEvent);
        CommunicationEvents.RemoveFactEvent.RemoveListener(DeleteFactEvent);
        _OnDisable();
    }
    protected virtual void _OnDisable() { }
    #endregion UnityMethods

    public void HiglightFactEvent(string fid, FactMaterials tmpState)
    {
        if (fid == URI)
            HighlightAs(tmpState);
    }
    public virtual void HighlightAs(FactMaterials tmpState) { }

    public void DeleteFactEvent(Fact fact)
    {
        if (fact.Id == URI)
            _DeleteFactEvent(fact);
    }
    protected virtual void _DeleteFactEvent(Fact fact)
        => Destroy(gameObject);

    protected virtual void FactUpdated() { }
}