﻿using Newtonsoft.Json;
using REST_JSON_API;
using System.Collections;
using System.Collections.Generic;
using UnityEngine;

/// <summary>
/// Base-class for 1D-Facts
/// </summary>
public abstract class AbstractLineFact : FactWrappedCRTP<AbstractLineFact>
{
    //protected override object GetCompiledValue() 
    //    => new Tuple<Vector3, Vector3>(Point1.Point, Point2.Point);

    /// @{ <summary>
    /// One <see cref="Fact.Id">Id</see> of two <see cref="PointFact"/> defining <see cref="Dir"/>.
    /// </summary>
    public string Pid1, Pid2;
    /// @}

    [JsonIgnore]
    public PointFact Point1 { get => (PointFact)FactRecorder.AllFacts[Pid1]; }
    [JsonIgnore]
    public PointFact Point2 { get => (PointFact)FactRecorder.AllFacts[Pid2]; }

    /// <summary> Distance between <see cref="AbstractLineFact.Pid1"/> and <see cref="AbstractLineFact.Pid2"/></summary>
    [JsonIgnore]
    public float Distance;

    /// <summary>
    /// Normalized Direction from <see cref="Pid1"/> to <see cref="Pid2"/>.
    /// </summary>
    public Vector3 Dir;

    /// <summary>
    /// \copydoc Fact.Fact()
    /// </summary>
    protected AbstractLineFact() : base()
    {
        Pid1 = null;
        Pid2 = null;
        Dir = Vector3.zero;
    }

    /// <summary>
    /// Standard Constructor
    /// </summary>
    /// <param name="pid1">sets <see cref="AbstractLineFact.Pid1"/></param>
    /// <param name="pid2">sets <see cref="AbstractLineFact.Pid2"/></param>
    protected AbstractLineFact(string pid1, string pid2) : base()
    {
        set_public_members(pid1, pid2);
    }

    /// <summary>
    /// Bypasses initialization of new MMT %Fact by using existend URI, _which is not checked for existence_.
    /// </summary>
    /// <param name="pid1">sets <see cref="Pid1"/></param>
    /// <param name="pid2">sets <see cref="Pid2"/></param>
    /// <param name="backendURI">MMT URI</param>
    protected AbstractLineFact(string pid1, string pid2, SOMDoc _ServerDefinition) : base()
    {
        this.ServerDefinition = _ServerDefinition;
        set_public_members(pid1, pid2);
    }

    /// <summary>
    /// Initiates <see cref="Pid1"/>, <see cref="Pid2"/>, <see cref="Dir"/>
    /// </summary>
    /// <param name="pid1">sets <see cref="Pid1"/></param>
    /// <param name="pid2">sets <see cref="Pid2"/></param>
    private void set_public_members(string pid1, string pid2)
    {
        this.Pid1 = pid1;
        this.Pid2 = pid2;
        Vector3 diff = (Point1.Point - Point2.Point);
        this.Dir = diff.normalized;
        this.Distance = diff.magnitude;
    }

    protected override void RecalculateTransform()
    {
        Position = Vector3.Lerp(Point1.Point, Point2.Point, 0.5f);
        Rotation = Quaternion.LookRotation(Dir, Vector3.up);
        LocalScale = new Vector3(1, 1, Distance);
    }

    /// \copydoc Fact.hasDependentFacts
    public override bool HasDependentFacts => true;

    /// \copydoc Fact.getDependentFactIds
    protected override string[] GetDependentFactIds()
        => new string[] { Pid1, Pid2 };
}

/// <summary>
/// Implements CRTP for <see cref="AbstractLineFact"/>; Escalates constructors;
/// </summary>
/// <typeparam name="T">class, which inherits from AbstractLineFactWrappedCRTP</typeparam>
public abstract class AbstractLineFactWrappedCRTP<T> : AbstractLineFact where T : AbstractLineFactWrappedCRTP<T>
{
    /// <summary>\copydoc Fact.Fact</summary>
    protected AbstractLineFactWrappedCRTP() : base() { }

    /// <summary>\copydoc AbstractLineFact.AbstractLineFact(string, string, FactOrganizer)</summary>
    protected AbstractLineFactWrappedCRTP(string pid1, string pid2) : base(pid1, pid2) { }

    /// <summary>\copydoc AbstractLineFact.AbstractLineFact(string, string, string, FactOrganizer)</summary>
    protected AbstractLineFactWrappedCRTP(string pid1, string pid2, SOMDoc _ServerDefinition) : base(pid1, pid2, _ServerDefinition) { }

    /// \copydoc Fact.Equivalent(Fact, Fact)
    protected override bool EquivalentWrapped(AbstractLineFact f1, AbstractLineFact f2)
        => EquivalentWrapped((T)f1, (T)f2);

    /// <summary>CRTP step of <see cref="EquivalentWrapped(AbstractLineFact, AbstractLineFact)"/></summary>
    protected abstract bool EquivalentWrapped(T f1, T f2);
}

/// <summary>
/// Line within 3D Space of finite length
/// </summary>
public class LineFact : AbstractLineFactWrappedCRTP<LineFact>
{
    /// \copydoc Fact.s_type
    [JsonProperty]
    protected static new string s_type = "LineFact";

    /// <summary> \copydoc Fact.Fact </summary>
    public LineFact() : base() { }

    /// <summary> \copydoc AbstractLineFact.AbstractLineFact(string, string, string, FactOrganizer) </summary>
    public LineFact(string pid1, string pid2, SOMDoc _ServerDefinition) : base(pid1, pid2, _ServerDefinition) { }

    /// <summary> \copydoc AbstractLineFact.AbstractLineFact(string, string, FactOrganizer) </summary>
    public LineFact(string pid1, string pid2) : base(pid1, pid2) { }

    public override MMTFact MakeMMTDeclaration()
    {
        SOMDoc valueTp = new OMS(MMTConstants.RealLit);
        OMLIT<float> value = new OMLIT<float>(Distance);

        return new MMTValueFact(_LastLabel, Defines(), valueTp, value);
    }

    public override SOMDoc Defines()
        => new OMA(
                new OMS(MMTConstants.Metric),
                new[] {
                    new OMS(Pid1),
                    new OMS(Pid2)
                }
            );

    /// \copydoc Fact.parseFact(ScrollFact)
    public new static IEnumerator parseFact(List<Fact> ret, MMTFact fact)
    {
        string pointAUri = ((OMS)((OMA)((MMTValueFact)fact).lhs).arguments[0]).uri;
        string pointBUri = ((OMS)((OMA)((MMTValueFact)fact).lhs).arguments[1]).uri;

        if (!FactRecorder.AllFacts.ContainsKey(pointAUri)
         || !FactRecorder.AllFacts.ContainsKey(pointBUri))
            yield break;

        ret.Add(new LineFact(pointAUri, pointBUri, fact.@ref));
    }

    /// \copydoc Fact.generateLabel
    protected override string generateLabel(FactRecorder name_space)
        => "[" + Point1.GetLabel(name_space) + Point2.GetLabel(name_space) + "]";

    /// \copydoc Fact.Equivalent(Fact, Fact)
    protected override bool EquivalentWrapped(LineFact f1, LineFact f2)
        => DependentFactsEquivalent(f1, f2);

    protected override Fact _ReInitializeMe(Dictionary<string, string> old_to_new)
        => new LineFact(old_to_new[this.Pid1], old_to_new[this.Pid2]);
}

/// <summary>
/// Ray within 3D Space of infinite length
/// </summary>
public class RayFact : AbstractLineFactWrappedCRTP<RayFact>
{
    /// \copydoc Fact.s_type
    [JsonProperty]
    protected static new string s_type = "RayFact";

    /// <summary> \copydoc Fact.Fact </summary>
    public RayFact() : base() { }

    /// <summary> \copydoc AbstractLineFact.AbstractLineFact(string, string, string, FactOrganizer) </summary>
    public RayFact(string pid1, string pid2, SOMDoc _ServerDefinition) : base(pid1, pid2, _ServerDefinition) { }

    /// <summary> \copydoc AbstractLineFact.AbstractLineFact(string, string, FactOrganizer) </summary>
    public RayFact(string pid1, string pid2) : base(pid1, pid2) { }

    public override MMTFact MakeMMTDeclaration()
    {
        SOMDoc type = new OMS(MMTConstants.LineType);

        return new MMTGeneralFact(_LastLabel, type, Defines());
    }

    public override SOMDoc Defines()
        => new OMA(
                new OMS(MMTConstants.LineOf),
                new[] {
                    new OMS(Pid1),
                    new OMS(Pid2)
            });

    /// \copydoc Fact.parseFact(ScrollFact)
    public new static IEnumerator parseFact(List<Fact> ret, MMTFact fact)
    {
        if (((MMTGeneralFact)fact).defines is not OMA defines)
            yield break;

        string pointAUri = ((OMS)defines.arguments[0]).uri;
        string pointBUri = ((OMS)defines.arguments[1]).uri;

        if (!FactRecorder.AllFacts.ContainsKey(pointAUri)
         || !FactRecorder.AllFacts.ContainsKey(pointBUri))
            yield break;

        ret.Add(new RayFact(pointAUri, pointBUri, fact.@ref));
    }

    protected override void RecalculateTransform()
    {
        base.RecalculateTransform();
        LocalScale = new Vector3(1, 1, 2048);
    }

    /// \copydoc Fact.generateLabel
    protected override string generateLabel(FactRecorder name_space)
    {
        // TODO this string is too large to properly depict on scrolls. 
        // return "]" + Point1.Label + Point2.Label + "[";
        return Point1.GetLabel(name_space) + Point2.GetLabel(name_space);
    }

    /// \copydoc Fact.Equivalent(Fact, Fact)
    protected override bool EquivalentWrapped(RayFact f1, RayFact f2)
    {
        if (!Math3d.IsApproximatelyParallel(f1.Dir, f2.Dir))
            return false;

        return Math3d.IsPointApproximatelyOnLine(f1.Point1.Point, f1.Dir, f2.Point1.Point)
            && Math3d.IsPointApproximatelyOnLine(f1.Point1.Point, f1.Dir, f2.Point2.Point);
    }

    protected override Fact _ReInitializeMe(Dictionary<string, string> old_to_new)
        => new RayFact(old_to_new[this.Pid1], old_to_new[this.Pid2]);
}
