using MoreLinq.Extensions;
using REST_JSON_API;
using System;
using System.Collections.Generic;
using System.Linq;
using UnityEngine;

public class CanonBallProblemCalculator3D
{
    private int Dim_G, Dim_A, Dim_B;
    private int MaxIter;

    private float Bounce;

    private FactRecorder FactOrganizer;

    private Vector3 StartPos;
    private Vector3 StartVel;
    private Vector3 Gravity;

    private List<QuadFact> Walls;
    public string Result_MovementFunc_Id;
    public List<string> Result_ArgsFunc_Id;
    public List<string> Result_FuncCall_Id;

    public struct Wall
    {
        public LineFact Topology;
        public float Bounce;

        public Wall(LineFact Topology, float Bounce)
        {
            this.Bounce = Bounce;
            this.Topology = Topology;
        }
    }

    public CanonBallProblemCalculator3D(List<QuadFact> walls, Vector3 starPos, Vector3 starVec, Vector3 gravity, float Bounce, int dim_G, int dim_A, int dim_B, FactRecorder factOrganizer, int maxIter = 32)
    {
        Walls = walls;
        Dim_G = dim_G;
        Dim_A = dim_A;
        Dim_B = dim_B;
        MaxIter = maxIter;
        StartPos = starPos;
        StartVel = starVec;
        Gravity = gravity;
        FactOrganizer = factOrganizer;
        this.Bounce = Bounce;

        Compute();
    }

    private void Compute()
    {
        SOMDoc BuildOMA_XVA()
        {   //Pos = Pos + Vel * t + 0.5 * g * t**2
            return new FUN(new FUN.Param[]
                {
                    new("Pos", new OMS(MMTConstants.TYPE_TO_OMS[typeof(Vector3)])),
                    new("Vel", new OMS(MMTConstants.TYPE_TO_OMS[typeof(Vector3)])),
                    new("Acc", new OMS(MMTConstants.TYPE_TO_OMS[typeof(Vector3)])),
                    new("t", new OMS(MMTConstants.TYPE_TO_OMS[typeof(float)])),
                },
                new OMA(
                    new OMS(MMTConstants.AddRealLit),
                    new SOMDoc[] {
                        new OMV("Pos"),
                        new OMA(
                            new OMS(MMTConstants.AddRealLit),
                            new[] {
                                new OMA(
                                    new OMS(MMTConstants.TimesRealLit),
                                    new[] {
                                        new OMV("Vel"),
                                        new OMV("t"),
                                    }
                                ),
                                new OMA(
                                    new OMS(MMTConstants.TimesRealLit),
                                    new SOMDoc[] {
                                        new OMLIT<float>(0.5f),
                                        new OMA(
                                            new OMS(MMTConstants.TimesRealLit),
                                            new SOMDoc[] {
                                                new OMV("Acc"),
                                                new OMA(
                                                    new OMS(MMTConstants.TimesRealLit),
                                                    new[] {
                                                        new OMV("t"),
                                                        new OMV("t"),
            }),}),}),}),}));
        }

        SOMDoc BuildOMAPath(Vector3 Pos, Vector3 Vel, float last_t)
        {   //t -> [Pos, Vel, g, t]
            return new FUN(new FUN.Param[]
                {
                    new("t", new OMS(MMTConstants.TYPE_TO_OMS[typeof(float)])),
                },
                new OMA(
                    new OMS(MMTConstants.MakeObjectArray),
                    new SOMDoc[] {
                        SOMDoc.MakeVector3(Pos),
                        SOMDoc.MakeVector3(Vel),
                        SOMDoc.MakeVector3(Gravity),
                        new OMA(
                            new OMS(MMTConstants.AddRealLit),
                            new SOMDoc[]{
                                new OMV("t"),
                                new OMA(
                                    new OMS(MMTConstants.MinusRealLit),
                                    new[]{ new OMLIT<float>(last_t) }
            )})}));
        }

        //Pos = Pos + Vel * t + 0.5 * g * t**2
        Vector3 UpdatePos(Vector3 Pos, Vector3 Vel, float t)
            => Pos + Vel * t + 0.5f * t * t * Gravity;

        //Vel = Vel + g * t
        Vector3 UpdateVel(Vector3 Vel, float t)
            => Vel + Gravity * t;

        // p1 + v1 * T + 0.5 * g1 * T^2 = o1 + s1 * s + t1 * t,
        // p2 + v2 * T + 0.5 * g2 * T^2 = o2 + s2 * s + t2 * t,
        // p3 + v3 * T + 0.5 * g3 * T^2 = o3 + s3 * s + t3 * t,

        // aT^2 + b*T + c = 0
        // T = (-b  sqrt(b ^ 2 - 4ac)) / (2a)
        //    Where:
        //        Any perumtation {1,2,3} is valid;
        //        Beware of Singularities in X,Y
        //
        //        X = ( s1 * t2 - s2 * t1 ) / ( s3 * t2 - s2 * t3 )
        //        Y = (-t1/t2 + X * t3/t2)
        //
        //        a = 0.5 *
        //           (G1 + Y * G2 - X * G3)
        //        b = V1 + Y * V2 - X * V3
        //        c = P1 + Y * P2 - X * P3
        //          -(O1 + Y * O2 - X * O3)
        //
        // see also: https://math.stackexchange.com/questions/866331/numerically-stable-algorithm-for-solving-the-quadratic-equation-when-a-is-very
        (float, float) SolveForTime(Vector3 Pos, Vector3 Vel, QuadFact Top)
        {
            Vector3
                s = Top.Tangents,
                t = Top.AltTangents;

            bool alt_sol = t[Dim_B] == 0 && s[Dim_B] == 0;
            if (t[Dim_B] == 0
             && (!alt_sol || t[Dim_G] == 0))
            {
                t = Top.Tangents;
                s = Top.AltTangents;
            }

            int sol_dim_A = 0,
                sol_dim_B = 0,
                sol_dim_C = 0;

            double X = 0, Y = 0;

            bool success = false;
            foreach (IList<int> perm in Enumerable.Range(0, 3).Permutations())
            {
                sol_dim_A = perm[0];
                sol_dim_B = perm[1];
                sol_dim_C = perm[2];

                X = (s[sol_dim_A] * t[sol_dim_B] - s[sol_dim_B] * t[sol_dim_A])
                  / (s[sol_dim_C] * t[sol_dim_B] - s[sol_dim_B] * t[sol_dim_C]);
                Y = -t[sol_dim_A] / t[sol_dim_B] + t[sol_dim_C] / t[sol_dim_B] * X;

                if (!(double.IsNaN(X) || double.IsInfinity(X)
                   || double.IsNaN(Y) || double.IsInfinity(Y)))
                {
                    success = true;
                    break;
                }
            }

            if (!success)
                return (-1, -1);

            double
                a = 0.5 * _Reduce(Gravity),
                b = _Reduce(Vel),
                c = _Reduce(Pos) - _Reduce(Top.Points[0].Point);

            if (Math.Abs(a) < Math3d.vectorPrecission)
                return Math.Abs(b) < Math3d.vectorPrecission
                    ? (-1, -1) // unconstrained
                    : ((float)(-c / b), -1);
            else if (Math.Abs(c) < Math3d.vectorPrecission) // not neccessary iff using double midnight
                return (0, (float)(-b / a));

            float
                t1 = (float)((-b - Math.Sign(b) * Math.Sqrt(b * b - 4 * a * c)) / (2 * a)),
                t2 = (float)(c / (a * t1)); // stable version, a.k.a. Viete's formula

            return (t1, t2);

            double _Reduce(Vector3 vec)
                => vec[sol_dim_A] + Y * vec[sol_dim_B] - X * vec[sol_dim_C];
        };

        bool HitOnPlane(Vector3 Pos, Vector3 Vel, QuadFact Top, float t)
        {
            Vector3 NewPos = UpdatePos(Pos, Vel, t);
            int dir = Math.Sign(Vector3.Dot(Vector3.Cross(Top.Normal, Top.Tangents), Top.AltTangents));

            for (int i = 0; i < Top.Points.Length; i++)
            {
                Vector3 ortho_edge =
                    Vector3.Cross(Top.Normal, Top.Points[(i + 1) % Top.Points.Length].Point - Top.Points[i].Point);
                Vector3 ToNewPos = NewPos - Top.Points[i].Point;

                if (dir != Math.Sign(Vector3.Dot(ortho_edge, ToNewPos)))
                    return false;
            }

            return true;
        };

        var Result_Movement = new FunctionFact(BuildOMA_XVA());
        Result_MovementFunc_Id = FactOrganizer.Add(Result_Movement, out _, false, null, null);

        Vector3 PlaneNorm = Vector3.zero;
        PlaneNorm[Dim_G] = 1;

        Vector3 pos = StartPos;
        Vector3 vel = StartVel;

        Result_ArgsFunc_Id = new();
        Result_FuncCall_Id = new();

        float last_t = 0f;
        for (int i = 0; i < MaxIter; i++)
        {
            pos = UpdatePos(pos, vel, (float)Math3d.vectorPrecission); //minimum step

            var hits = Walls
                .Select(w => (w, tarr: SolveForTime(pos, vel, w)))
                .SelectMany(wt => new[] { (wt.w, t: wt.tarr.Item1), (wt.w, t: wt.tarr.Item2) })
                .Where(wt => wt.t >= 0 && HitOnPlane(pos, vel, wt.w, wt.t))
                .OrderBy(wt => wt.t)
                .ToArray();

            (QuadFact next_Wall, float next_t) = hits.Length > 0
                ? hits[0]
                : (default(QuadFact), float.PositiveInfinity);

            Result_ArgsFunc_Id.Add(
                FactOrganizer.Add(
                    new FunctionFact(BuildOMAPath(pos, vel, last_t - (float)Math3d.vectorPrecission)),
                    out _, true, null, null));

            Result_FuncCall_Id.Add(
                FactOrganizer.Add(
                    new FunctionCallFact(Result_MovementFunc_Id, Result_ArgsFunc_Id.Last(), (-(float)Math3d.vectorPrecission + last_t, next_t + last_t)),
                    out _, true, null, null));

            if (hits.Length == 0)
                break;

            pos = UpdatePos(pos, vel, next_t);
            Vector3 vel_at_bounce = UpdateVel(vel, next_t);

            vel = Bounce * Vector3.Reflect(vel_at_bounce, next_Wall.Normal);
            last_t += next_t;
        }

        return;
    }
}
